@@ -5,20 +5,21 @@
 # Author            : FFunction
 # License           : BSD License
 # -----------------------------------------------------------------------------
-# Creation date     : 2013-JUL-15
-# Last modification : 2014-DEC-24
+# Creation date     : 14-Jul-2013
+# Last modification : 30-Dec-2014
 # -----------------------------------------------------------------------------
 
-import re, sys
-from   libparsing import Grammar, Token, Word, Rule, Group, Condition, Procedure, Reference, AbstractProcessor
+import re, os, sys, argparse, json
+from   libparsing import Grammar, Token, Word, Rule, Group, Condition, Procedure, Reference, AbstractProcessor, TreeWriter
 
 try:
 	import reporter
 except ImportError:
 	reporter = None
 
-VERSION = "0.0.5"
+VERSION = "0.0.8"
 LICENSE = "http://ffctn.com/doc/licenses/bsd"
+
 __doc__ = """
 Processor for the PythonicCSS language. This module use a PEG-based parsing
 engine <http://github.com/sebastien/parsing>, which sadly has an important
@@ -74,9 +75,9 @@ def grammar(g=None):
 	g.token   ("NUMBER",           "-?(0x)?[0-9]+(\.[0-9]+)?")
 	g.token   ("ATTRIBUTE",        "[a-zA-Z\-_][a-zA-Z0-9\-_]*")
 	g.token   ("ATTRIBUTE_VALUE",  "\"[^\"]*\"|'[^']*'|[^,\]]+")
-	g.token   ("SELECTOR_SUFFIX",  ":[\-a-z][a-z0-9\-]*(\([0-9]+\))?")
-	g.token   ("SELECTION_OPERATOR", "\>|[ ]+")
-	g.word    ("INCLUDE",             "%include")
+	g.token   ("SELECTOR_SUFFIX",  "::?[\-a-z][a-z0-9\-]*(\([^\)]+\))?")
+	g.token   ("SELECTION_OPERATOR", "\>|\+|[ ]+")
+	g.word    ("INCLUDE",          "%include")
 	g.word    ("COLON",            ":")
 	g.word    ("DOT",              ".")
 	g.word    ("LP",               "(")
@@ -84,6 +85,7 @@ def grammar(g=None):
 	g.word    ("RP",               ")")
 	g.word    ("SELF",             "&")
 	g.word    ("COMMA",            ",")
+	g.word    ("SEMICOLON",        ";")
 	g.word    ("EQUAL",            "=")
 	g.word    ("LSBRACKET",        "[")
 	g.word    ("RSBRACKET",        "]")
@@ -91,8 +93,8 @@ def grammar(g=None):
 	g.token   ("PATH",             "\"[^\"]+\"|'[^']'|[^\s\n]+")
 	g.token   ("PERCENTAGE",       "\d+(\.\d+)?%")
 	g.token   ("STRING_SQ",        "'((\\\\'|[^'\\n])*)'")
-	g.token   ("STRING_DQ",        "\"((\\\\\"|[^\"\\n])*)\"")
-	g.token   ("STRING_UQ",        "[^\s\n\*\\-\+;\(\)\[\]]+")
+	g.token   ("STRING_DQ",        "\"((\\\\\"|[^\"\n])*)\"")
+	g.token   ("STRING_UQ",        "[^\s\n\*\+;\(\)\[\]]+")
 	g.token   ("INFIX_OPERATOR",   "[\-\+\*\/]")
 
 	g.token   ("NODE",             "\*|([a-zA-Z][\-_a-zA-Z0-9\-]*)")
@@ -108,7 +110,7 @@ def grammar(g=None):
 	g.token   ("COLOR_NAME",       "[a-z][a-z0-9\-]*")
 	g.token   ("COLOR_HEX",        "\#([A-Fa-f0-9][A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?[A-Fa-f0-9]?([A-Fa-f0-9][A-Fa-f0-9])?)")
 	g.token   ("COLOR_RGB",        "rgba?\((\s*\d+\s*,\s*\d+\s*,\s*\d+\s*(,\s*\d+(\.\d+)?\s*)?)\)")
-	g.token   ("CSS_PROPERTY",    "[a-z][a-z0-9\-]*")
+	g.token   ("CSS_PROPERTY",    "[\-a-z][\-a-z0-9]*")
 	g.token   ("SPECIAL_NAME",     "@[A-Za-z][A-Za-z0-9\_\-]*")
 	g.token   ("SPECIAL_FILTER",   "\[[^\]]+\]")
 
@@ -123,7 +125,7 @@ def grammar(g=None):
 	g.rule      ("Attribute",        s.ATTRIBUTE._as("name"), g.arule(s.EQUAL, s.ATTRIBUTE_VALUE).optional()._as("value"))
 	g.rule      ("Attributes",       s.LSBRACKET, s.Attribute._as("head"), g.arule(s.COMMA, s.Attribute).zeroOrMore()._as("tail"), s.RSBRACKET)
 
-	g.rule      ("Selector",         g.agroup(s.SELF, s.NODE).optional()._as("scope"), s.NODE_ID.optional()._as("nid"), s.NODE_CLASS.zeroOrMore()._as("nclass"), s.Attributes.optional()._as("attributes"), s.SELECTOR_SUFFIX.zeroOrMore()._as("suffix"))
+	g.rule      ("Selector",         g.agroup(s.SELF, s.NODE).optional()._as("scope"), s.NODE_ID.optional()._as("nid"), s.NODE_CLASS.zeroOrMore()._as("nclass"), s.Attributes.zeroOrMore()._as("attributes"), s.SELECTOR_SUFFIX.zeroOrMore()._as("suffix"))
 	g.rule      ("SelectorNarrower", s.SELECTION_OPERATOR._as("op"), s.Selector._as("sel"))
 
 	g.rule      ("Selection",        g.agroup(s.Selector)._as("head"), s.SelectorNarrower.zeroOrMore()._as("tail"))
@@ -135,7 +137,7 @@ def grammar(g=None):
 
 	g.group     ("Suffixes")
 	g.rule      ("Number",           s.NUMBER._as("value"), s.UNIT.optional()._as("unit"))
-	g.group     ("String",           s.STRING_UQ, s.STRING_SQ, s.STRING_DQ)
+	g.group     ("String",           s.STRING_SQ, s.STRING_DQ, s.STRING_UQ)
 	g.group     ("Value",            s.Number, s.COLOR_HEX, s.COLOR_RGB, s.REFERENCE, s.String)
 	g.rule      ("Parameters",       s.VARIABLE_NAME, g.arule(s.COMMA, s.VARIABLE_NAME).zeroOrMore())
 	g.rule      ("Arguments",        s.Value, g.arule(s.COMMA, s.Value).zeroOrMore())
@@ -162,24 +164,22 @@ def grammar(g=None):
 	# OPERATIONS
 	# =========================================================================
 
-	g.rule      ("Assignment",       s.CSS_PROPERTY._as("name"), s.COLON, s.Expression.oneOrMore()._as("values"), s.IMPORTANT.optional()._as("important"))
+	g.rule      ("Assignment",       s.CSS_PROPERTY._as("name"), s.COLON, s.Expression.oneOrMore()._as("values"), s.IMPORTANT.optional()._as("important"), s.SEMICOLON.optional())
 	g.rule      ("MacroInvocation",  s.MACRO_NAME,   s.LP, s.Arguments.optional(), s.RP)
 
 	# =========================================================================
 	# BLOCK STRUCTURE
 	# =========================================================================
 
-	g.group("Code")
 	# NOTE: If would be good to sort this out and allow memoization for some
 	# of the failures. A good idea would be to append the indentation value to
 	# the caching key.
 	# .processMemoizationKey(lambda _,c:_ + ":" + c.getVariables().get("requiredIndent", 0))
 	g.rule("Statement",     s.CheckIndent._as("indent"), g.agroup(s.Assignment, s.MacroInvocation, s.COMMENT), s.EOL).disableFailMemoize()
-	g.rule("Block",         s.CheckIndent._as("indent"), g.agroup(s.PERCENTAGE, s.SelectionList)._as("selector"), s.COLON.optional(), s.EOL, s.Indent, s.Code.zeroOrMore()._as("code"), s.Dedent).disableFailMemoize()
-	s.Code.set(s.Block, s.Statement).disableFailMemoize()
+	g.rule("Block",         s.CheckIndent._as("indent"), g.agroup(s.PERCENTAGE, s.SelectionList)._as("selector"), s.COLON.optional(), s.EOL, s.Indent, s.Statement.zeroOrMore()._as("code"), s.Dedent).disableFailMemoize()
 
 	g.rule    ("SpecialDeclaration",   s.CheckIndent, s.SPECIAL_NAME, s.SPECIAL_FILTER.optional(), s.Parameters.optional(), s.COLON)
-	g.rule    ("SpecialBlock",         s.CheckIndent, s.SpecialDeclaration, s.EOL, s.Indent, s.Code.zeroOrMore(), s.Dedent).disableFailMemoize()
+	g.rule    ("SpecialBlock",         s.CheckIndent, s.SpecialDeclaration, s.EOL, s.Indent, s.Statement.zeroOrMore(), s.Dedent).disableFailMemoize()
 
 	# =========================================================================
 	# AXIOM
@@ -217,9 +217,10 @@ class Processor(AbstractProcessor):
 
 	RGB        = None
 
-	RE_SPACES = re.compile("\s+")
-
-	COLOR_PROPERTIES = (
+	RE_SPACES            = re.compile("\s+")
+	RE_UNQUOTED          = re.compile("\!?[\./\w\d_-]+")
+	FIX_STRING_DELIMITER = "\\\"\\\""
+	COLOR_PROPERTIES     = (
 		"background",
 		"background-color",
 		"color",
@@ -269,11 +270,19 @@ class Processor(AbstractProcessor):
 	OPERATIONS = {
 		"+" : lambda a,b,u:[a[0] + b[0], a[1] or b[1] or u],
 		"-" : lambda a,b,u:[a[0] - b[0], a[1] or b[1] or u],
-		"*" : lambda a,b,u:[a[0] * b[0], a[1] or b[1] or u],
-		"/" : lambda a,b,u:[a[0] / b[0], a[1] or b[1] or u],
+		"*" : lambda a,b,u:[float(a[0]) * b[0], a[1] or b[1] or u],
+		"/" : lambda a,b,u:[float(a[0]) / b[0], a[1] or b[1] or u],
 		"%" : lambda a,b,u:[a[0] % b[0], a[1] or b[1] or u],
 	}
 
+	OPERATOR_PRIORITY = {
+		"+" : 0,
+		"-" : 0,
+		"*" : 1,
+		"/" : 1,
+		"%" : 1,
+	}
+
 	@classmethod
 	def IsColorProperty( cls, name ):
 		return name in cls.COLOR_PROPERTIES
@@ -335,9 +344,9 @@ class Processor(AbstractProcessor):
 			if self.IsColorProperty(name) and v[1] == "S":
 				# We have a color name as a string in a color property, we expand it
 				return (self.ColorFromName(v[0]) or v[0], "C")
-			elif v[1] == "S" and name in self.PREFIXABLE_VALUES_PROPERTIES:
-				# We're in a property that refernces prexialbe properties
-				if prefix and v[0] in self.PREFIXABLE_PROPERTIES:
+			elif v[1] == "S":
+				if name in self.PREFIXABLE_VALUES_PROPERTIES and prefix and v[0] in self.PREFIXABLE_PROPERTIES:
+					# We're in a property that refernces prexialbe properties
 					return (prefix + v[0], v[1])
 				else:
 					return v
@@ -364,6 +373,8 @@ class Processor(AbstractProcessor):
 			# TODO: Should detect the scope type and apply the corresponding method
 			r += "({0})".format(",".join((self._valueAsString(_) for _ in args or [])))
 			return (r, None)
+		elif e[0] == "(":
+			return self.evaluate(e[1])
 		else:
 			raise ProcessingException("Evaluate not implemented for: {0} in {1}".format(e, name))
 
@@ -412,10 +423,10 @@ class Processor(AbstractProcessor):
 		return match.group()
 
 	def onSTRING_DQ(self, match ):
-		return match.group(1)
+		return self._stringEscapeFix(match.group(1))
 
 	def onSTRING_SQ(self, match ):
-		return match.group(1)
+		return self._stringEscapeFix(match.group(1))
 
 	def onSTRING_UQ(self, match ):
 		return match.group()
@@ -434,10 +445,13 @@ class Processor(AbstractProcessor):
 		return [self.defaultProcess(match[0])] + [self.process(_[1]) for _ in self.process(match[1])]
 
 	def onArguments( self, match ):
-		return [self.evaluate(_) for _ in [self.process(match[0])] + [self.process(_[1]) for _ in self.process(match[1])]]
+		m0 = self.process(match[0])
+		m1 = self.process(match[1])
+		p = [m0] + ([_[1] for _ in m1] if m1 else [])
+		return [self.evaluate(_) for _ in p]
 
 	def onInvocation( self, match, method, arguments ):
-		return ["I", None, method[1], arguments]
+		return ["I", None, method and method[1] or None, arguments]
 
 	def onInfixOperation( self, match ):
 		op   = self.process(match[0])
@@ -452,7 +466,7 @@ class Processor(AbstractProcessor):
 		child          = match[0]
 		grand_children = child.children()
 		result         = self.process(child)
-		result         = result[1] if len(result) == 3 else result
+		result         = ["(", result[1]] if len(result) == 3 else result
 		return result
 
 	def onExpression( self, match ):
@@ -465,10 +479,23 @@ class Processor(AbstractProcessor):
 			elif suffix[0] == "I":
 				suffix[1] = res
 			res = suffix
+		# We rework operator precedence here
+		if res[0] == "O" and res[3][0] == "O":
+			# We're in this situation
+			# ['O', '*', ['V', (4, None)], ['O', '+', ['V', (10, None)], ['V', (5, None)]]]
+			# while we ant
+			# ['O', '+', ['O', '*', ['V', (4, None)], ['V', (10, None)]], ['V', (5, None)]]]
+			op  = res[1]
+			rop = res[3][1]
+			if self.OPERATOR_PRIORITY[rop] < self.OPERATOR_PRIORITY[op]:
+				a = res[2]
+				b = res[3][2]
+				c = res[3][3]
+				res = ["O", rop, ["O", op, a, b], c]
 		return res
 
 	def onAttribute( self, match, name, value ):
-		return "[{0}{1}{2}]".format(name, value[0], value[1])
+		return "[{0}{1}{2}]".format(name, value[0] if value else "", value[1] if value else "")
 
 	def onAttributes( self, match, head, tail ):
 		assert not tail
@@ -479,10 +506,11 @@ class Processor(AbstractProcessor):
 		"""Selectors are returned as tuples `(scope, id, class, attributes, suffix)`.
 		We need to keep this structure as we need to be able to expand the `&`
 		reference."""
-		scope  = scope[0] if scope else ""
-		nid    = nid if nid else ""
-		suffix = "".join(suffix) if suffix else ""
-		nclass = "".join(nclass) if nclass else ""
+		scope      = scope[0] if scope else ""
+		nid        = nid if nid else ""
+		suffix     = "".join(suffix) if suffix else ""
+		nclass     = "".join(nclass) if nclass else ""
+		attributes = "".join(attributes) if attributes else ""
 		if (scope or nid or nclass or attributes or suffix):
 			return [scope, nid, nclass, attributes or "", suffix]
 		else:
@@ -490,7 +518,8 @@ class Processor(AbstractProcessor):
 
 	def onSelectorNarrower( self, match, op, sel ):
 		"""Returns a `(op, selector)` couple."""
-		op = op and (op.strip() + " ") or ""
+		if op:
+			op = op.strip() or " "
 		return (op, sel) if op or sel else None
 
 	def onSelection( self, match, head, tail ):
@@ -498,17 +527,22 @@ class Processor(AbstractProcessor):
 		>   [[('div', '', '', '', ''), '> ', ('label', '', '', '', '')]]
 		>   ---SELECTOR------------   OP   --SELECTOR---------------
 		"""
+		if not head and not tail: return None
 		if head:
 			res = head
 			if tail:
 				for narrower in tail:
+					if not narrower: continue
 					res.extend(narrower)
 			return res
 		else:
 			res = []
 			for i, v in enumerate(tail):
 				if i == 0:
-					res.append(v[1])
+					if v[0] == " ":
+						res.append(v[1])
+					else:
+						res.extend(v)
 				else:
 					res.extend(v)
 			return res
@@ -616,9 +650,9 @@ class Processor(AbstractProcessor):
 						# This is a tricky operation, but we get the very first
 						# selector of the given scope, which starts with an &,
 						# merge it with the most specific part of the current
-						# scopes and prepend the rest of th full scope.
-						merged = self._mergeScopes(scope[0], full_scope[-1])
-						res.append(full_scope[0:-1] + [merged])
+						# scopes and prepend the rest of th fulle scope.
+						merged = [self._mergeScopes(scope[0], full_scope[-1])] + scope[1:]
+						res.append(full_scope[0:-1] + merged)
 				else:
 					for full_scope in parent_scopes:
 						res.append(full_scope + [" "] + scope)
@@ -630,7 +664,7 @@ class Processor(AbstractProcessor):
 		"""Helper function for _mergeScopes"""
 		if not a: return b
 		if not b: return a
-		return a + b
+		return b + a
 
 	def _mergeScopes( self, a, b):
 		# Merges the contents of scope A and B
@@ -642,31 +676,51 @@ class Processor(AbstractProcessor):
 	def _selectionAsString( self, selection ):
 		return "".join(self._scopeAsString(_) if isinstance(_, list) else _ for _ in selection if _) if selection else ""
 
+	def _stringEscapeFix( self, text ):
+		# NOTE: CleverCSS had some trouble with the empty content string, which
+		# lead us to write content: "\"\"/\"\"". This escapes that
+		if text.startswith(self.FIX_STRING_DELIMITER) and text.endswith(self.FIX_STRING_DELIMITER):
+			return text[len(self.FIX_STRING_DELIMITER):-len(self.FIX_STRING_DELIMITER)]
+		else:
+			return text
+
 	def _valueAsString( self, value ):
 		v, u = value ; u = u or ""
 		if   u == "C":
-			if len(v) == 3:
+			if type(v) in (str, unicode):
+				# If we have a string instead of a tuple, we pass it as-is
+				return v
+			elif len(v) == 3:
 				r, g, b = v
 				r = ("0" if r < 16 else "") + hex(r)[2:].upper()
 				g = ("0" if g < 16 else "") + hex(g)[2:].upper()
 				b = ("0" if b < 16 else "") + hex(b)[2:].upper()
 				return "#" + r + g + b
-			else:
+			elif len(v) == 4:
 				return "rgba({0},{1},{2},{3:0.2f})".format(*v)
+			else:
+				raise ProcessingException("Expected RGB triple, RGBA quadruple or string, got: {0} in {1}".format(v, value))
 		if   type(v) == int:
 			return "{0:d}{1}".format(v,u)
 		elif type(v) == float:
 			# We remove the trailing 0 to have the most compact representation
-			v = "{0:f}".format(v)
+			v = str(v)
 			while len(v) > 2 and v[-1] == "0" and v[-2] != ".":
 				v = v[0:-1]
+			if v.endswith(".0"):v = v[:-2]
 			return v + u
 		elif type(v) == str:
 			# FIXME: Proper escaping
-			return "{0:s}".format(v,u)
+			if self.RE_UNQUOTED.match(v):
+				return "{0:s}".format(v,u)
+			else:
+				return "{0:s}".format(json.dumps(v),u)
 		elif type(v) == unicode:
 			# FIXME: Proper escaping
-			return "{0:s}".format(v,u)
+			if self.RE_UNQUOTED.match(v):
+				return "{0:s}".format(v,u)
+			else:
+				return "{0:s}".format(json.dumps(v),u)
 		else:
 			raise ProcessingException("Value string conversion not implemented: {0}".format(value))
 
@@ -690,19 +744,38 @@ def run(args):
 	USAGE = "pythoniccss FILE..."
 	if reporter: reporter.install(reporter.StderrReporter())
 	if type(args) not in (type([]), type(())): args = [args]
-	from optparse import OptionParser
+	oparser = argparse.ArgumentParser(
+		prog        = os.path.basename(__file__.split(".")[0]),
+		description = "Compiles PythonicCSS files to CSS"
+	)
+	oparser.add_argument("files", metavar="FILE", type=str, nargs='+', help='The .pcss files to parse')
+	oparser.add_argument("--report",  dest="report",  action="store_true", default=False)
+	oparser.add_argument("--output",  type=str,  dest="output", default=None)
 	# We create the parse and register the options
-	oparser = OptionParser(prog="pythoniccss", description=__doc__,
-	usage=USAGE, version="PythonicCSS " + VERSION)
-	options, args = oparser.parse_args(args=args)
-	p = Processor(output=sys.stdout)
-	if not args:
+	args = oparser.parse_args(args=args)
+	p   = Processor(output=sys.stdout)
+	# p = TreeWriter(output=sys.stdout)
+	if not args.files:
 		sys.stderr.write(USAGE + "\n")
-	for a in args:
+	output = sys.stdout
+	if args.output: output = open(args.output, "w")
+	for a in args.files:
 		if reporter: reporter.info("Processing: {0}".format(a))
 		match = parse(a)
-		p.process(match)
-
+		if args.report:
+			output.write("Report for : {0}\n".format(a))
+			stats = match.stats()
+			stats.report(getGrammar(), output)
+		else:
+			if match.status() == "S":
+				p.process(match)
+			else:
+				msg = "Parsing of {0} failed at line:{1}".format(a, match.line())
+				reporter.error(msg)
+				reporter.error(match.textaround())
+				raise Exception("Parsing of {0} failed at line:{1}\n> {2}".format(a, match.line(), match.textaround()))
+	if args.output:
+		output.close()
 if __name__ == "__main__":
 	import sys
 	run(sys.argv[1:])
